# __init__.py

# Version of the combinatorial_tests package
__version__ = "0.0.4"

from comb_testing.tuple_set import TupleSet
from comb_testing.test_suite import TestSuite
from comb_testing.biased_algorithm import generate_biased_suite
import re


# function to generate a covering array based on input
# input follows formatting "Input as follows - \"#Levels^#Factors\" -
# put a space between each for multi-level covering:"
def generate_covering_array(re_input):
    match_list = re.findall("(\\d+)\\^(\\d+)", re_input)

    cover_arr = []
    lvl_offset = 0
    for match in match_list:
        levels = int(match[0])
        factors = int(match[1])

        for i in range(factors):
            row = []
            for j in range(levels):
                row.append(j + lvl_offset)

            lvl_offset += levels
            cover_arr.append(row)
    return cover_arr


# function to take a test suite and output contents to file
def test_suite_output(suite):
    file_out = open("aetg_output.txt", "w")

    file_out.write(str(len(suite)) + "\n\n")

    for candidate in suite:
        file_out.write(str(candidate) + "\n")

    file_out.close()


def check_valid_input(new_list):
    uniq = set()
    for fct in new_list:
        if len(fct) == 0:
            return False
        else:
            for lvl in fct:
                if lvl in uniq:
                    return False
                else:
                    uniq.add(lvl)
    return True


# function to run the greedy version of the combinatorial testing algorithm
# new_list is a 2D list where the outer layer represents the factors and the inner layer represents the levels
# strength represents the n of n-way coverage
def greedy_algorithm(new_list, strength, flag):
    if strength > len(new_list):
        print("Error: Coverage strength greater than factor count")
        return []
    if not check_valid_input(new_list):
        print("Error: Input list is invalid")
        return

    tuples = TupleSet(new_list, strength)
    tuples.n_way_recursion(0, (), 0)
    tuples.update_tuples()

    # generate a test suite
    suite = TestSuite(tuples)

    if flag == 1:
        return suite.generate_greedy_suite_size()
    elif flag == 2:
        tuples.generate_combos()
        return suite.generate_greedy_suite_speed()


# function to run the biased version of the combinatorial testing algorithm
# new_list and benefit_list 2D lists of equal size and depth which map a "benefit" value to each value in the main list
# exclusions is a list which contains pairs to be excluded from final test cases
def biased_algorithm(new_list, benefit_list, exclusions):
    if not check_valid_input(new_list):
        print("Error: Input list is invalid")
        return

    return generate_biased_suite(new_list, benefit_list, exclusions)



# TestSuite class within comb_testing package
# Copyright Andrew Ragland 2021

import itertools
import random


class TestSuite:
    def __init__(self, tuple_dict):
        self.tuples = tuple_dict
        self.covering_arr = tuple_dict.get_covering_arr()
        self.factor_count = len(self.covering_arr)
        self.suite = []
        self.factor_order = list(range(self.factor_count))

    def check_valid_candidate(self, candidate):
        for fct_ind in range(self.factor_count):
            if not candidate[fct_ind] in self.covering_arr[fct_ind]:
                return False
        return True

    # generate a test suite and return it using the greedy method
    def generate_greedy_suite_size(self):

        while not self.tuples.is_empty():

            curr_candidate = [-1] * self.factor_count

            random.shuffle(self.factor_order)
            potential_lvls = []
            best_count = -1
            curr_factor = 0

            # first factor
            factor_index = self.factor_order[curr_factor]

            for val in self.covering_arr[factor_index]:
                self.tuples.count_tuples_value(val, factor_index, 1, (val,), 0)
                if (count := self.tuples.get_tuple_count()) > best_count:
                    best_count = count
                    potential_lvls[:] = [val]
                elif count == best_count:
                    potential_lvls.append(val)

            curr_candidate[factor_index] = random.choice(potential_lvls)
            curr_factor += 1

            # remaining factors
            while curr_factor < self.factor_count:
                best_count = -1
                potential_lvls[:] = []
                factor_index = self.factor_order[curr_factor]

                for val in self.covering_arr[factor_index]:
                    curr_candidate[factor_index] = val
                    self.tuples.count_tuples_candidate(curr_candidate)
                    if (count := self.tuples.get_tuple_count()) > best_count:
                        best_count = count
                        potential_lvls[:] = [val]
                    elif count == best_count:
                        potential_lvls.append(val)

                curr_candidate[factor_index] = random.choice(potential_lvls)
                curr_factor += 1

            self.suite.append(curr_candidate)
            self.tuples.cover_tuples(curr_candidate)

        self.tuples.reset_tuples()
        return self.suite

    # generate a test suite by first generating all possible candidate rows and add to suite based on maximum coverage
    def generate_greedy_suite_speed(self):
        while not self.tuples.is_empty():
            counts = []
            for row in self.tuples.combos:
                self.tuples.count_tuples_candidate(row)
                counts.append(self.tuples.get_tuple_count())
            test = list(itertools.zip_longest(self.tuples.combos, counts))
            test = sorted(test, key=lambda x: x[1], reverse=True)

            for row in test:
                if self.check_valid_candidate(row[0]):
                    self.tuples.cover_tuples(row[0])
                    self.suite.append(row[0])
                    break

        return self.suite



# TupleSet class within comb_testing package
# Copyright Andrew Ragland 2021
import itertools


class TupleSet:

    def __init__(self, new_list, n):
        self.tuple_set = set()
        self.tuple_copy = set()
        self.covering_arr = new_list
        self.factor_count = len(self.covering_arr)
        self.covered = 0
        self.total_tuples = 0
        self.strength = n
        self.tuple_count = 0

        self.combos = []

    def generate_combos(self):
        self.combos = list(
            itertools.combinations([lvl for fct in self.covering_arr for lvl in fct], len(self.covering_arr)))

    def n_way_recursion(self, depth, t, f):
        if depth == self.strength:
            self.tuple_set.add(t)
        else:
            for fct in range(f, self.factor_count):
                for lvl in self.covering_arr[fct]:
                    nest_t = t + (lvl,)
                    self.n_way_recursion(depth + 1, nest_t, fct + 1)

    def update_tuples(self):
        self.total_tuples = len(self.tuple_set)
        self.tuple_copy = self.tuple_set.copy()

    def count_tuples_value(self, val, index, depth, t, f):
        if depth == self.strength:
            if t in self.tuple_set:
                self.tuple_count += 1
        else:
            for fct in range(f, len(self.covering_arr)):
                if val in self.covering_arr[fct]:
                    continue
                else:
                    if fct < index:
                        for lvl in self.covering_arr[fct]:
                            nest_t = (lvl,) + t
                            self.count_tuples_value(val, index, depth + 1, nest_t, fct + 1)
                    elif fct > index:
                        for lvl in self.covering_arr[fct]:
                            nest_t = t + (lvl,)
                            self.count_tuples_value(val, index, depth + 1, nest_t, fct + 1)

    # counter the number of tuples a specified candidate can cover
    def count_tuples_candidate(self, candidate):
        tuples = itertools.combinations(candidate, self.strength)
        for t in tuples:
            if t in self.tuple_set:
                self.tuple_count += 1

    # cover the tuples as found within a specified candidate
    def cover_tuples(self, candidate):
        tuples = itertools.combinations(candidate, self.strength)
        for t in tuples:
            if t in self.tuple_set:
                self.tuple_set.discard(t)
                self.covered += 1

    # once a test suite has covered all tuples, uncover all tuples such that another test suite can run
    def reset_tuples(self):
        self.tuple_set = self.tuple_copy.copy()
        self.covered = 0

    def get_total_tuples(self):
        return self.total_tuples

    def get_tuples_covered(self):
        return self.covered

    # after using count_tuples_value/_candidate, test suite gets count value and resets it for next count call
    def get_tuple_count(self):
        ret = self.tuple_count
        self.tuple_count = 0
        return ret

    def is_empty(self):
        return False if self.tuple_set else True

    def get_covering_arr(self):
        return self.covering_arr



# Biased Covering Array Algorithm - Andrew Ragland - 2021

import itertools
import random


# calculate a candidate's benefit value
def calc_candidate_benefit(tuples, candidate):
    can_benefit = 0
    tuple_count = 0
    pairs = list(itertools.combinations(candidate, 2))
    for pair in pairs:
        for tup in list(tuples.items()):
            if tup[0] == pair:
                can_benefit += tup[1]
                tuple_count += 1

    return can_benefit, tuple_count


# build a candidate out of a list of pairs
def build_candidate(cvr, pairs):
    candidate = [-1] * len(cvr)

    for pair in pairs:
        for fct in cvr:
            if pair[0] in fct:
                candidate[cvr.index(fct)] = pair[0]
            elif pair[1] in fct:
                candidate[cvr.index(fct)] = pair[1]

    return candidate


# perform pair generation using the recursive method
def two_way_recursion(depth, t, b, f, tuples, cvr, bft):
    if depth == 2:
        tuples.update({t: b})
    else:
        for fct in range(f, len(cvr)):
            for lvl in cvr[fct]:
                nest_t = t + (lvl,)
                nest_b = round(b * bft[fct][cvr[fct].index(lvl)], 2)
                two_way_recursion(depth + 1, nest_t, nest_b, fct + 1, tuples, cvr, bft)


def generate_biased_suite(covering_array, benefit_array, exclusion_array):
    cover_arr = covering_array
    benefit_arr = benefit_array
    exclusions = exclusion_array

    tuples = dict()

    test_suite = []

    two_way_recursion(0, (), 1, 0, tuples, cover_arr, benefit_arr)

    # exclusions - remove excluded pairs from the list of tuples
    if exclusions:
        for e in exclusions:
            tuples.pop(e)

    tuples = dict(sorted(tuples.items(), key=lambda item: item[1], reverse=True))

    while tuples:

        # store pairs where their benefit values are both maximum and equal
        eq_benefit_list = []
        eq_benefit_val = list(tuples.items())[0][1]

        factor_order = list(range(len(cover_arr)))
        random.shuffle(factor_order)

        for t in list(tuples.items()):
            if t[1] < eq_benefit_val:
                break
            else:
                eq_benefit_list.append(t[0])

        # random check
        init_tuple = random.choice(eq_benefit_list)

        init_can = build_candidate(cover_arr, [init_tuple])
        best_can = []
        for f in factor_order:
            max_benefit = 0
            max_count = 0
            if init_can[f] == -1:
                for val in cover_arr[f]:
                    init_can[f] = val
                    benefit, count = calc_candidate_benefit(tuples, init_can)
                    if benefit > max_benefit or count > max_count:
                        best_can = init_can.copy()
                        max_benefit = benefit
                        max_count = count

        # perform exclusions
        pairs = list(itertools.combinations(best_can, 2))

        final_can = []
        for p in pairs:
            if p in exclusions:
                # start exclusion process, determine pair indicies
                exc_benefit = 0
                exc_count = 0
                indices = [-1] * 2
                for i in range(len(cover_arr)):
                    if p[0] in cover_arr[i]:
                        indices[0] = i
                    elif p[1] in cover_arr[i]:
                        indices[1] = i

                for ind in indices:
                    exc_can = best_can.copy()
                    for val in cover_arr[ind]:
                        if val not in p:
                            exc_can[ind] = val
                            benefit, count = calc_candidate_benefit(tuples, exc_can)
                            if benefit > exc_benefit or count > exc_count:
                                final_can = exc_can.copy()
                                exc_benefit = benefit
                                exc_count = count
            else:
                final_can = best_can.copy()

        #
        final_pairs = list(itertools.combinations(final_can, 2))
        for p in final_pairs:
            for t in list(tuples.items()):
                if t[0] == p:
                    tuples.pop(t[0])
                    break
        if final_can:
            test_suite.append(final_can)

    return test_suite



import pathlib
from setuptools import setup


HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()


setup(
    name='combinatorial_tests_agragland',
    version='0.0.4',
    packages=["comb_testing"],
    py_modules=["test_suite", "tuple_set"],
    url='https://github.com/agragland/comb-test-library',
    license='MIT',
    author='Andrew Ragland',
    author_email='agragland00@gmail.com',
    description='Creates combinatorial test suites',
    long_description=README,
    long_description_content_type="text/markdown",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    python_requires='>=3.8'
)
